"""
basic functions: format, validate, display tracking
"""

import os
import re
from datetime import datetime

from PySide2 import QtCore
from obspy import UTCDateTime
import numpy as np
from sohstationviewer.view.util.enums import LogType


def validateFile(path2file, fileName):
    """
    Check if fileName given is a file and not info file
    :param path2file: str - absolute path to file
    :param fileName: str - name of the file
    :return: bool - True if pass checking, False if not.
    """

    if fileName.strip() == '.DS_Store' or fileName.startswith('._'):
        # skip mac's info file
        return False

    if not os.path.isfile(path2file):
        return False
    return True


@QtCore.Slot()
def displayTrackingInfo(trackingBox, text, type=LogType.INFO):
    """
    Display text in the given widget with different background and text colors
    :param trackingBox: QTextBrowser - widget to display tracking info
    :param text: str - info to be displayed
    :param type: str - (info/warning/error) type of info to be displayed in
        different color
    """

    if trackingBox is None:
        print(f"{type.name}: {text}")
        return

    msg = {'text': text}
    if type == LogType.ERROR:
        msg['color'] = 'white'
        msg['bgcolor'] = '#e46269'
    elif type == LogType.WARNING:
        msg['color'] = '#ffd966'
        msg['bgcolor'] = 'orange'
    else:
        msg['color'] = 'blue'
        msg['bgcolor'] = 'white'
    htmlText = """<body>
        <div style='color:%(color)s; background-color:%(bgcolor)s'>
            %(text)s
        </div>
        </body>"""
    trackingBox.setHtml(htmlText % msg)
    # parent.update()
    trackingBox.repaint()


def getDirSize(dir):
    """
    Get size of directory and size of file.
    :param dir: str - absolute path to directory
    :return:
        totalSize: int - total size of the directory
        totalFile: int - total file of the directory
    """

    totalSize = 0
    totalFile = 0
    for path, subdirs, files in os.walk(dir):
        for fileName in files:
            if not validateFile(os.path.join(path, fileName), fileName):
                continue
            fp = os.path.join(path, fileName)
            # print("file %s: %s" % (fp, os.path.getsize(fp)))
            totalSize += os.path.getsize(fp)
            totalFile += 1
    return totalSize, totalFile


def getTime6(timeStr):
    """
    Get time from 6 parts string.
    (year:day of year:hour:minute:second:millisecond)
    Ex: 01:251:09:41:35:656/ 2001:251:09:41:35:656
    in which year in the first part can be 2 digits or 6 digits
    :param timeStr: str - 6 part time string
    """
    year = timeStr.split(':')[0]
    if len(year) == 2:
        return getTime6_2y(timeStr)
    else:
        return getTime6_4y(timeStr)


def getTime6_2y(timeStr):
    """
    Get time from 6 parts string in which year has 2 digits.
    Ex: 01:251:09:41:35:656
    :param timeStr: str - 6 part time string with 2 digits for year
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(22, "0")
    time = datetime.strptime(timeStr, "%y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime6_4y(timeStr):
    """
    Get time from 6 parts string in which year has 4 digits.
    Ex: 2001:251:09:41:35:656
    :param timeStr: str - 6 part time string with 4 digits for year
    """
    # pad 0 so the last part has 6 digits to match with the format str
    timeStr = timeStr.ljust(24, "0")
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S:%f")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year


def getTime4(timeStr, trackingYear, yAdded):
    """
    Get time from 4 parts string. (day of year:hour:minute:second)
    Ex: 253:19:41:42
    :param timeStr: str - time string
    :param trackingYear: int - year that has been detected
    :param yAdded: bool - flag to tell if year has been plussed 1 or not
    :return:
        + utcTime.timestamp: float: epoch time
        + time.year: int: year
        + yAdded: bool - flag to tell if year has been plussed 1 or not
    """
    if not yAdded:
        # first day => move to next year
        doy = int(timeStr.split(':')[0])
        if doy == 1:
            trackingYear += 1
            yAdded = True
    timeStr = f'{str(trackingYear)}:{timeStr}'
    time = datetime.strptime(timeStr, "%Y:%j:%H:%M:%S")
    utcTime = UTCDateTime(time)
    return utcTime.timestamp, time.year, yAdded


def getVal(text):
    """
    Get the value part of a string with non-number substring following.
    :param text: str - value string including unit
    :return: value part including +/-, remove str that follows - float
    """
    REVal = '^\+?\-?[0-9]+\.?[0-9]?'            # noqa: W605
    return float(re.search(REVal, text).group())


def isBinaryStr(text):
    """
    :param text: str: text to check
    :return: bool - True if text is a binary string or False if not
    """

    return lambda b: bool(b.translate(None, text))


def rtnPattern(text, upper=False):
    """
    This function is from logpeek's rtnPattern.
    return routine pattern of the string with:
     + 0 for digit
     + a for lowercase
     + A for upper case
     + remain special character
    :param text: str - text to get format
    :return rtn: str - routine pattern of the string
    """
    rtn = ""
    for c in text:
        if c.isdigit():
            rtn += "0"
        elif c.isupper():
            rtn += "A"
        elif c.islower():
            rtn += "a"
        else:
            rtn += c
    # So the A/a chars will always be A, so the caller knows what to look for
    if upper is True:
        return rtn.upper()
    return rtn


def fmti(Value):
    """
    This function is from logpeek's fmti
    Given Value will be convert to a string integer with thousand separators
    :param Value: str - string of value with unit
    :return NewValue: str - new value with no unit and with thousand separators
    """
    Value = int(Value)
    if Value > -1000 and Value < 1000:
        return str(Value)
    Value = str(Value)
    NewValue = ""
    # There'll never be a + sign.
    if Value[0] == "-":
        Offset = 1
    else:
        Offset = 0
    CountDigits = 0
    for i in range(len(Value) - 1, -1 + Offset, -1):
        NewValue = Value[i] + NewValue
        CountDigits += 1
        if CountDigits == 3 and i != 0:
            NewValue = "," + NewValue
            CountDigits = 0
    if Offset != 0:
        if NewValue.startswith(","):
            NewValue = NewValue[1:]
        NewValue = Value[0] + NewValue
    return NewValue


def apply_convert_factor(c_data, convert_factor):
    """
    convertFactor = 150mV/count = 150V/1000count
    => unit data * convertFactor= data *150/1000 V

    :param c_data: dict - data of the channel which includes down-sampled
        data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
        soh_data[key][chan_id]
    :param convert_factor: float - convertFactor field retrieved from
        db table Channels for this channel
    """
    c_data['data'] = np.multiply(c_data['data'], [convert_factor])
