"""
MSeed object to hold and process MSeed data
"""

import os
from pathlib import Path

from sohstationviewer.conf import constants
from sohstationviewer.controller.util import validateFile
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.handling_data import (
    readWaveformMSeed, squash_gaps, checkWFChan, sortData, readSOHTrace,
)
from sohstationviewer.model.mseed.from_mseedpeek.mseed_header import readHdrs
from sohstationviewer.view.util.enums import LogType


class MSeed(DataTypeModel):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # channels: set - available channels got from files among
        # requested channels
        self.channels = set()

        # nets: set - available nets got from files
        self.nets = set()

        # stats: set - available stats got from files
        self.stats = set()

        """
        A file can have more than one experiment which only one of them should
        be correct as selected by user (handling_data.py - read_soh_mseed())
        netsProbInFile is the dictionary with key is the tuple of all nets
        in a file, value is the selected net selected by user.
        """
        self.netsProbInFile = {}

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.read_soh_and_index_waveform(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selectedKey = self.selectStaID()

        if self.selectedKey is None:
            raise ThreadStopped()
        if len(self.reqWFChans) != 0:
            self.readWFFiles(self.selectedKey)

    def read_soh_and_index_waveform(self, folder):
        """
            + read waveform data for filename associate with time range
            + read soh_stream for unmerged traces
            + merge soh_stream into SOH data and mass position data
            + squash gaps from different soh channels to one
        :param folder: str - absolute path to data set folder
        """
        self.waveformData, soh_stream = self.index_waveform(folder)
        self.SOHData, self.massPosData, self.gaps = self.merge_soh_streams(
            soh_stream)

    def index_waveform(self, folder):
        """
        :param folder: str - absolute path to data set folder
        :return waveformData: a dict for waveform data including
        {sta_id: {'filesInfo': {chan_id: [{filepath,time range,trace's info}]}
                 'readData': { chan_id: {samplerate
                         and preset traces_info which is a list of {info and
                         np.memmap's file path of trace data and time}
                  }
        }}
            (use np.memmap because data are too large to be merged and
                processed)
        :return soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        """
        soh_streams = {}
        waveform_data = {}
        net_stat_prop = {}
        stat_prop = {}
        chan_prop = {}
        count = 0

        for path, sub_dirs, files in os.walk(folder):
            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)
                if not validateFile(path2file, file_name):
                    continue
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} file headers/ SOH files', LogType.INFO)

                ret = readHdrs(
                    path2file, file_name, soh_streams, self.logData,
                    self.reqSOHChans, self.reqWFChans,
                    self.netsProbInFile, self.trackInfo)
                if ret is None:
                    continue

                self.nets.update(ret['nets'])
                self.stats.add(ret['stats'][0])

                if len(ret['stats']) > 1:
                    k = tuple(ret['stats'])
                    if k not in stat_prop:
                        stat_prop[k] = []
                    stat_prop[k].append(file_name)
                else:
                    if len(ret['netStats']) > 1:
                        k = tuple(ret['netStats'])
                        if k not in net_stat_prop:
                            net_stat_prop[k] = []
                        net_stat_prop[k].append(file_name)
                if len(ret['chanIDs']) > 1:
                    k = tuple(ret['chanIDs'])
                    if k not in chan_prop:
                        chan_prop[k] = []
                    chan_prop[k].append(file_name)
                self.channels.update(ret['chanIDs'])
                if 'read' not in ret:
                    # not waveform
                    continue

                # TODO currently for waveform not read the file
                #  with 2 stations, 2 channels
                # correctly. Use the first one in the file only
                # When have time will dig more to this
                sta_id = ret['stats'][0]
                chan_id = ret['chanIDs'][0]
                if sta_id not in waveform_data:
                    waveform_data[sta_id] = {"filesInfo": {},
                                             "readData": {}}
                    self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
                if chan_id not in waveform_data[sta_id]["filesInfo"]:
                    waveform_data[sta_id]["filesInfo"][chan_id] = []
                    waveform_data[sta_id]["readData"][chan_id] = {
                        "samplerate": ret['spr'],
                        "tracesInfo": []}
                waveform_data[sta_id]['filesInfo'][chan_id].append(ret)

        if len(stat_prop) > 0:
            errmsg = (f"More than one stations in a file: {stat_prop}. "
                      f"Will use the first one.")
            self.trackInfo(errmsg, LogType.ERROR)
        if len(net_stat_prop) > 0:
            errmsg = "More than one netIDs in a file: %s" % net_stat_prop
            self.trackInfo(errmsg, LogType.WARNING)
        if len(chan_prop) > 0:
            errmsg = (f"More than one channels in a file: {chan_prop} "
                      f"\nThis is a CRITICAL ERROR.")
            self.trackInfo(errmsg, LogType.ERROR)
        return waveform_data, soh_streams

    def merge_soh_streams(self, soh_streams):
        """
        :param soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        if there is more than one network for a station, user can either choose
        data for one network for that station, or combine all networks related
        to that station into one data set (by changing network name to the
        selected network then merge again)
        :return masspos_data: dict - merged data for chan_id with 'VM' prefixed
            (structure in DataTypeModel.__init__.massPosData)
        :return soh_data: dict - merged data for other soh chan_ids
            (structure in DataTypeModel.__init__SOHData)
        :return gaps: dict - start and end of gaps from all merged soh streams
            are squashed to the largest gaps
            (structure in DataTypeModel.__init__gaps)
        """

        soh_data = {}
        masspos_data = {}
        gaps = {}
        all_gaps = []
        for sta_id in soh_streams:
            self.stats.add(sta_id)
            self.dataTime[sta_id] = [constants.HIGHEST_INT, 0]
            soh_data[sta_id] = {}
            masspos_data[sta_id] = {}
            for chan_id in soh_streams[sta_id]:
                self.channels.add(chan_id)
                stream = soh_streams[sta_id][chan_id]

                stream.merge()

                tr = stream[0]
                if len(stream) > 1:
                    nets = [tr.stats['network'].strip() for tr in stream]
                    nets += [f"Combine to {n}" for n in nets]
                    msg = (f"There are more than one net for sta {sta_id}.\n"
                           "Please select one or combine all to one.")
                    self.pause_signal.emit(msg, nets)
                    self.pause()
                    sel_net = nets[self.pause_response]
                    if "Combine" not in sel_net:
                        tr = [tr for tr in stream
                              if tr.stats['network'] == sel_net][0]
                        self.nets.add(sel_net)
                    else:
                        for tr in stream:
                            sel_net = sel_net.replace("Combine to ", "")
                            tr.stats['network'] = sel_net
                            self.nets.add(sel_net)
                        stream.merge()
                        tr = stream[0]

                gaps_in_stream = stream.get_gaps()
                all_gaps += [[g[4].timestamp, g[5].timestamp]
                             for g in gaps_in_stream]
                trace_info = readSOHTrace(tr)
                if chan_id.startswith('VM'):
                    masspos_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                else:
                    soh_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                self.dataTime[sta_id][0] = min(trace_info['startTmEpoch'],
                                               self.dataTime[sta_id][0])
                self.dataTime[sta_id][1] = max(trace_info['endTmEpoch'],
                                               self.dataTime[sta_id][1])

            gaps[sta_id] = squash_gaps(all_gaps)
        return soh_data, masspos_data, gaps

    def selectStaID(self):
        """
        :return selectedStaID: str - the selected station id from self.stats
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """

        stats = list(self.stats)
        selectedStaID = stats[0]
        if len(stats) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, stats)
            self.pause()
            selectedStaID = stats[self.pause_response]

        self.trackInfo(f'Select Station {selectedStaID}', LogType.INFO)
        return selectedStaID

    def readWFFiles(self, staID):
        """
        From filesInfo, read all waveform data of requested waveform channels
        for given sta_id, in the selected time (from Main Window) to add
        to waveformData[sat_id]['readData'][chan_id]['tracesInfo'].
        'tracesInfo' includes list of traces' info dict in startTmEpoch order.
        The keys of a trace's info dict:
         {
            'samplerate': sample rate of the data,
            'startTmEpoch': start epoch time of the trace,
            'endTmEpoch': end epoch time of the trace,
            'size': size of data,
            'times_f': np.memmap's file path for time
            'data_f': np.memmap's file path for data
        }
        :param staID: str - station ID
        """

        count = 0
        for chanID in self.waveformData[staID]['filesInfo']:
            # check chanID
            hasChan = checkWFChan(chanID, self.reqWFChans)
            if not hasChan:
                continue
            tracesInfo = self.waveformData[staID][
                'readData'][chanID]['tracesInfo']

            for fileInfo in self.waveformData[staID]['filesInfo'][chanID]:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped
                # file have been read
                if fileInfo['read']:
                    continue

                # check time
                hasData = False

                if ((self.readStart <= fileInfo['startEpoch'] <= self.readEnd) or       # noqa: E501
                        (self.readStart <= fileInfo['endEpoch'] <= self.readEnd)):      # noqa: E501
                    hasData = True
                if not hasData:
                    continue
                readWaveformMSeed(fileInfo['path2file'], fileInfo['fileName'],
                                  staID, chanID, tracesInfo,
                                  self.dataTime[staID], self.tmpDir)
                fileInfo['read'] = True
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} waveform files', LogType.INFO)

        sortData(self.waveformData)
