
from sohstationviewer.conf import constants
from sohstationviewer.controller.util import (
    getTime6, getTime4, getVal, rtnPattern)
from sohstationviewer.view.util.enums import LogType


class LogInfo():

    def __init__(self, parent, trackInfo, logText, key, reqDSs,
                 isLogFile=False):
        """
        Help to extract channel data from LogText which include SOH Message and
        Event message.
        :param parent: reftek object - that calls LogInfo
        :param trackInfo: function - to track data processing
        :param logText: str - SOH and Event messages in time order
        :param key: (str, str) - ID of the data set including unitID and expNo
        :param reqDSs: list of str - requested data stream ID
        :param isLogFile: bool - flag indicate if this is a log file
        """
        self.parent = parent
        self.trackInfo = trackInfo
        self.logText = logText
        self.key = key
        self.unitID, self.expNo = key
        self.reqDSs = reqDSs
        self.isLogFile = isLogFile
        """
        trackYear to add year to time since year time not include year after
        header.
        yearChanged: if doy=1, trackYear is added 1, yAdded is marked
        this has happened to only added 1 once.
        """
        self.yAdded = False
        self.trackYear = 0
        if self.unitID >= "9000":
            self.model = "RT130"
        else:
            self.model = "72A"
        self.maxEpoch = 0
        self.minEpoch = constants.HIGHEST_INT
        self.chans = self.parent.SOHData[self.key]
        self.CPUVers = set()
        self.GPSVers = set()
        self.extractInfo()

    def readEVT(self, line):
        """
        Read EVT info from line for a specific datastream (DS)
        :param line: str - a line of evt message
        :return epoch: float: epoch time of message
        :return DS: int: index of data stream
            for epoch, using trigger time (TT) if available or
            first sample time (FST) if available, otherwise return 0, 0
        minEpoch and maxEpoch are updated.
        """
        # Ex: DAS: 0108 EV: 2632 DS: 2 FST = 2001:253:15:13:59:768
        #       TT =2001:253:15:13:59:768 NS: 144005 SPS: 40 ETO: 0
        parts = line.split()
        DS = int(parts[5])
        if DS not in self.parent.reqDSs:
            return (0, 0)
        try:
            if parts[8].startswith("00:000"):
                if parts[11].startswith("00:000"):
                    return -1, 0
                epoch, _ = getTime6(parts[11])
            else:
                epoch, _ = getTime6(parts[8])
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        if epoch > 0:
            self.minEpoch = min(epoch, self.minEpoch)
            self.maxEpoch = max(epoch, self.maxEpoch)
        else:
            return 0, 0
        return epoch, DS

    def readSHHeader(self, line):
        """
        :param line: str - a line of evt message
        :return epoch: float - time for state of health header
        minEpoch and maxEpoch are updated.
        yAdded is reset to false to allow adding 1 to trackYear
        If different unitID is detected, give warning and skip reading.
        """
        # Ex: State of Health  01:251:09:41:35:656   ST: 0108
        parts = line.split()
        try:
            epoch, self.trackYear = getTime6(parts[3])
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        self.yAdded = False         # reset yAdded
        self.minEpoch = min(epoch, self.minEpoch)
        self.maxEpoch = max(epoch, self.maxEpoch)
        unitID = parts[5].strip()
        if unitID != self.unitID:
            msg = ("The State Of Health messages for DAS %s were being "
                   "read, but now there is information for DAS %s in "
                   "the same State Of Health messages.\n"
                   "Skip reading State Of Health." % (self.unitID, unitID))
            self.trackInfo(msg, LogType.ERROR)
            False
        return epoch

    def simpleRead(self, line):
        """
        Read parts and epoch from an SOH line
        :param line: str - a line of evt message
        :return parts: list of str - parts of line with space delim
        :return epoch: float - time when info is recorded
        maxEpoch is updated with the epoch time.
        """
        # Ex: 186:21:41:35 <content>
        parts = line.split()
        try:
            epoch, self.trackYear, self.yAdded = getTime4(
                parts[0], self.trackYear, self.yAdded)
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        self.maxEpoch = max(epoch, self.maxEpoch)
        return parts, epoch

    def readIntClockPhaseErr(self, line):
        """
        Read internal clock phase error
        :param line: str - a line of evt message
        :return epoch: float - time when info is recorded
        :return error: float - time of ICP error in microseconds
        """
        # Ex: 253:19:41:42 INTERNAL CLOCK PHASE ERROR OF 4823 USECONDS
        ret = self.simpleRead(line)
        if not ret:
            return False
        parts, epoch = ret
        error = float(parts[-2])
        if parts[-1].startswith("SEC"):
            error *= 1000000.0
        return epoch, error

    def readBatTemBkup(self, line):
        """
        Read battery voltage, temperature, backup voltage
        :param line: str - a line of evt message
        :return epoch: float - time when info is recorded
        :return volts: float - battery voltage.
        :return temp: float - temperature of battery in Celsius
        :return bkupV: float - backup voltage.
            Available for RT130. For 72A, 0.0 is assigned
        """
        # 72A's:
        # Ex: 186:14:33:58 BATTERY VOLTAGE = 13.6V, TEMPERATURE = 26C
        # RT130:
        # Ex: 186:14:33:58 BATTERY VOLTAGE = 13.6V, TEMPERATURE = 26C,
        #    BACKUP = 3.3V
        parts, epoch = self.simpleRead(line)
        try:
            volts = getVal(parts[4])
            temp = getVal(parts[7])
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        if self.model == "RT130":
            bkupV = getVal(parts[10])
        else:
            bkupV = 0.0
        return epoch, volts, temp, bkupV

    def readDiskUsage(self, line):
        """
        Read disk usage
        :param line: str - a line of evt message
        :return epoch: float - time when info is recorded
        :return disk: int -  disk number
        :return val: int - memory used in disk
            Available for RT130. For 72A, 0 is assigned
        """
        # RT130:
        # Ex: 186:14:33:58 DISK 1: USED: 89744 AVAIL:...
        # Ex: 186:14:33:58 DISK 2* USED: 89744 AVAIL:...
        parts, epoch = self.simpleRead(line)
        try:
            disk = getVal(parts[2])
            val = getVal(parts[4])
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        return epoch, disk, val

    def readDPS_ClockDiff(self, line):
        """
        Read DPS clock difference
        :param line: str - a line of evt message
        :return epoch: float - time when info is recorded
        :return total: float - total difference time in milliseconds
        """
        # Ex: 245:07:41:45 DSP CLOCK DIFFERENCE: 0 SECS AND -989 MSECS
        parts, epoch = self.simpleRead(line)
        try:
            secs = getVal(parts[4])
            msecs = getVal(parts[-2])
        except AttributeError:
            self.parent.processingLog.append(line, LogType.ERROR)
            return False
        total = abs(secs) * 1000.0 + abs(msecs)
        if secs < 0.0 or msecs < 0.0:
            total *= -1.0
        return epoch, total

    def readDefs(self, line):
        """
        Read definitions' time. Currently, only read Station Channel Definition
        Based on user requested, may use ["STATION", "DATA", "CALIBRATION"]
        :param line: str - a line of evt message
        :return epoch: float - time of the definition
        """
        # Ex: Station Channel Definition   01:330:19:24:42:978    ST: 7095
        # Ex: Data Stream Definition   01:330:19:24:42:978    ST: 7095
        # Ex: Calibration Definition   01:330:19:24:42:978    ST: 7095
        # Ex: STATION CHANNEL DEFINITION  2020:066:19:00:56:000   ST: 92E9
        parts = line.split()
        # Lines from a .log file may be just the first time the parameters were
        # saved until the DAS is reset, so if this is a log file then use the
        # current SOH time for plotting the points, instead of what is in the
        # message line.
        if parts[0] in ["STATION"]:
            if self.isLogFile is False:
                try:
                    epoch, _ = getTime6(parts[-3])
                except AttributeError:
                    self.parent.processingLog.append(line, LogType.ERROR)
                    return False
            else:
                epoch = self.maxEpoch
        else:
            return False
        # These will not be allowed to set TimeMin since the RT130's save
        # multiple  copies of the parameters in the log files (one per day),
        # but they all have the date/timestamp of the first time the parameters
        # were written (unless a new copy is written because the parameters
        # were changed, of course).
        return epoch

    def readCPUVer(self, line):
        """
        Read version of CPU software
        :param line: str - a line of evt message
        :return CPUVer: str - version of CPU software
        """
        # Ex: 341:22:05:41 CPU SOFTWARE V03.00H  (72A and older 130 FW)
        # Ex: 341:22:05:41 REF TEK 130  (no version number at all)
        # Ex: 341:22:05:41 Ref Tek 130  2.8.8S (2007:163)
        # Ex: 341:22:05:41 CPU SOFTWARE V 3.0.0 (2009:152)
        parts = line.split()
        if parts[1].startswith("CPU"):
            CPUVer = " ".join(parts[3:])
        elif parts[1].upper().startswith("REF"):
            # There may not be any version info in this line:
            # Ex: REF TEK 130...then nothing
            # but accessing non-existant InParts doesn't matter.
            # CPUVer will just be "".
            CPUVer = " ".join(parts[4:])
        return CPUVer

    def readGPSVer(self, line):
        """
        Read version of GPS firmware
        :param line: str - a line of evt message
        :return GPSVer: str - version of GPS firmware
        """
        parts = line.split()
        verParts = [p.strip() for p in parts]
        if "GPS FIRMWARE VERSION:" in line:
            # 130 Ex: 291:19:54:29 GPS FIRMWARE VERSION: GPS 16-HVS Ver. 3.20
            GPSVer = ' '.join(verParts[4:])
        else:
            # 72A: Ex: 159:15:41:05 GPS: V03.30    (17JAN2000)
            GPSVer = " ".join(verParts[2:])
        return GPSVer

    def addChanInfo(self, chan_id, t, d, idx):
        """
        Add information to field orgTrace of channel
        {
            unitID: serial number of device
            expNo: experiment number
            times: list of epoch time
            data: list of channel's values
            logIdx: list of indexes of SOH message line corresponding to
                time and data
        }
        :param chan_id: str - ID of channel
        :param t: float - epoch time of data
        :param d: float - value of data
        :param idx: int - index of SOH message line
        """

        if chan_id not in self.chans:
            self.chans[chan_id] = {}
            self.chans[chan_id]['orgTrace'] = {
                'unitID': self.unitID,
                'expNo': self.expNo,
                'times': [],
                'data': [],
                'logIdx': []}
        self.chans[chan_id]['orgTrace']['times'].append(t)
        self.chans[chan_id]['orgTrace']['data'].append(d)
        self.chans[chan_id]['orgTrace']['logIdx'].append(idx)

    def extractInfo(self):
        """
        Extract data from each line of log string to add to
        SOH channels's orgTrace using addChanInfo()
        """
        lines = [ln.strip() for ln in self.logText.splitlines()]
        sohEpoch = 0

        for idx, line in enumerate(lines):

            if line == '':
                continue
            line = line.upper()
            if 'FST' in line:
                ret = self.readEVT(line)
                if ret:
                    epoch, DS = ret
                    if DS in self.reqDSs:
                        if epoch > 0:
                            chanName = 'Event DS%s' % DS
                            self.addChanInfo(chanName, epoch, 1, idx)
                        elif epoch == 0:
                            self.parent.processingLog.append(
                                line, LogType.WARNING)
                        else:
                            self.parent.processingLog.append(
                                line, LogType.ERROR)

            elif line.startswith("STATE OF HEALTH"):
                epoch = self.readSHHeader(line)
                if epoch:
                    self.addChanInfo('SOH/Data Def', epoch, 1, idx)
            elif "DEFINITION" in line:
                epoch = self.readDefs(line)
                if epoch:
                    self.addChanInfo('SOH/Data Def', epoch, 0, idx)

            if "INTERNAL CLOCK PHASE ERROR" in line:
                ret = self.readIntClockPhaseErr(line)
                if ret:
                    epoch, err = ret
                    self.addChanInfo('Clk Phase Err', epoch, err, idx)

            elif "POSSIBLE DISCREPANCY" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Discrepancies', epoch, 1, idx)

            elif "BATTERY VOLTAGE" in line:
                ret = self.readBatTemBkup(line)
                if ret:
                    epoch, volts, temp, bkupV = ret
                    self.addChanInfo('Battery Volt', epoch, volts, idx)
                    self.addChanInfo('DAS Temp', epoch, temp, idx)
                    # bkupV: x<3, 3<=x<3.3, >3.3
                    self.addChanInfo('Backup Volt', epoch, bkupV, idx)

            elif all(x in line for x in ['DISK', 'USED']):
                ret = self.readDiskUsage(line)
                if ret:
                    epoch, disk, val = ret
                    self.addChanInfo(f'Disk Usage{int(disk)}', epoch, val, idx)

            elif "AUTO DUMP CALLED" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Dump Called/Comp', epoch, 1, idx)
            elif "AUTO DUMP COMPLETE" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Dump Called/Comp', epoch, 0, idx)

            elif "DSP CLOCK SET" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Jerks/DSP Sets', epoch, 1, idx)
            elif 'JERK' in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Jerks/DSP Sets', epoch, 0, idx)

            elif "DPS clock diff" in line:
                ret = self.readDPS_ClockDiff()
                if ret:
                    epoch, total = ret
                    self.addChanInfo('DPS Clock Diff', epoch, total, idx)

            elif "ACQUISITION STARTED" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('ACQ On/Off', epoch, 1, idx)
            elif "ACQUISITION STOPPED" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('ACQ On/Off', epoch, 0, idx)

            elif "NETWORK LAYER IS UP" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Net Up/Down', epoch, 0, idx)
            elif "NETWORK LAYER IS DOWN" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Net Up/Down', epoch, 1, idx)

            elif "MASS RE-CENTER" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Mass Re-center', epoch, 0, idx)

            elif any(x in line for x in ["SYSTEM RESET", "FORCE RESET"]):
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Reset/Power-up', epoch, 0, idx)
            elif "SYSTEM POWERUP" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Reset/Power-up', epoch, 1, idx)

            # ================= GPS ==================================
            elif "EXTERNAL CLOCK IS UNLOCKED" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS Lk/Unlk', epoch, 0, idx)
            elif "EXTERNAL CLOCK IS LOCKED" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS Lk/Unlk', epoch, 1, idx)

            elif any(x in line for x in ["EXTERNAL CLOCK POWER IS TURNED ON",
                                         "EXTERNAL CLOCK WAKEUP",
                                         "GPS: POWER IS TURNED ON",
                                         "GPS: POWER IS CONTINUOUS"]
                     ):
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS On/Off/Err', epoch, 1, idx)
            elif any(x in line for x in ["EXTERNAL CLOCK POWER IS TURNED OFF",
                                         "EXTERNAL CLOCK SLEEP",
                                         "GPS: POWER IS TURNED OFF",
                                         "NO EXTERNAL CLOCK"]
                     ):
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS On/Off/Err', epoch, 0, idx)
            elif "EXTERNAL CLOCK ERROR" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS On/Off/Err', epoch, -1, idx)

            elif "EXTERNAL CLOCK CYCLE" in line:
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('GPS Clock Power', epoch, 1, idx)

            # ================= VERSIONS ==============================
            elif any(x in line for x in ["REF TEK", "CPU SOFTWARE"]):
                CPUVer = self.readCPUVer(line)
                self.CPUVers.add(CPUVer)

            elif any(x in line for x in ["GPS FIRMWARE VERSION:", "GPS: V"]):
                GPSVer = self.readGPSVer(line)
                self.GPSVers.add(GPSVer)

            # ================= ERROR/WARNING =========================
            elif "ERROR:" in line:
                # These lines are generated by programs like ref2segy and do
                # not have any time associated with them so just use whatever
                # time was last in Time.
                self.addChanInfo('Error/Warning', sohEpoch, -2, idx)
            elif any(x in line for x in ["NO VALID DISK", "FAILED"]):
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Error/Warning', epoch, -1, idx)

            elif "WARNING" in line:
                # Warings come in lines with the time from the RT130 and lines
                # created by programs without times.
                if rtnPattern(line[:12]) == "000:00:00:00":
                    epoch = self.simpleRead(line)[1]
                    if epoch:
                        self.addChanInfo('Error/Warning', epoch, 0, idx)
                else:
                    # Just use whatever time was last in Time.
                    # The 2 means that the dots will be a little different.
                    self.addChanInfo('Error/Warning', sohEpoch, 1, idx)

            elif (all(x in line for x in ["BAD", "MAKER"]) or
                  any(x in line for x in ["ERTFS", "IDE BUSY"])):
                epoch = self.simpleRead(line)[1]
                if epoch:
                    self.addChanInfo('Error/Warning', epoch, 0, idx)
