"""
RT130 object to hold and process RefTek data
"""

import os
from pathlib import Path
import numpy as np

from obspy.core import Stream

from sohstationviewer.model.reftek.from_rt2ms import (
    core, soh_packet, packet)
from sohstationviewer.model.reftek.logInfo import LogInfo
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.handling_data import (
    readWaveformReftek, squash_gaps, sortData, readMPTrace, readText)

from sohstationviewer.conf import constants

from sohstationviewer.controller.util import validateFile

from sohstationviewer.view.util.enums import LogType


class RT130(DataTypeModel):
    """
    read and process reftek file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwarg):
        self.EH = {}
        super().__init__(*args, **kwarg)
        self.keys = set()
        self.reqDSs = self.reqWFChans
        self.massPosStream = {}

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.readSOH_indexWaveform(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selectedKey = self.selectKey()
        if self.selectedKey is None:
            raise ThreadStopped()

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        if len(self.reqWFChans) != 0:
            self.readWFFiles(self.selectedKey)

    def readSOH_indexWaveform(self, folder):
        """
        Loop all files in dir to read for soh data, mass position data and
            index waveform data with filename and corresponding time range
        :param folder: str - absolute path to data set folder
        """
        count = 0
        for path, subdirs, files in os.walk(folder):
            for fileName in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()
                path2file = Path(path).joinpath(fileName)
                if not validateFile(path2file, fileName):
                    continue
                if not self.readReftek130(path2file):
                    readText(path2file, fileName, self.logData['TEXT'])
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} file headers/ SOH files', LogType.INFO)

        self.combineData()

    def selectKey(self):
        """
        :return selectedKey: (str, str) -
            (device's serial number, experiment_number)
            the selected keys from self.keys.
        + If there is only one key, return it.
        + If there is more than one, show all keys, let user choose one to
        return.
        """
        self.keys = sorted(list(self.keys))
        selectedKey = self.keys[0]
        if len(self.keys) > 1:
            msg = ("There are more than one keys in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, self.keys)
            self.pause()
            selectedKey = self.keys[self.pause_response]

        self.trackInfo(f'Select Key {selectedKey}', LogType.INFO)
        return selectedKey

    def readWFFiles(self, key):
        """
        :param key: (str, str) -
            (device's serial number, experiment_number)
            the selected keys from self.keys.
        From filesInfo, read all waveform data of requested waveform DS
            channels for given sta_id, in the selected time (from Main Window)
            to append to DataTypeModel.__init__.
            waveformData[sat_id]['readData'][chan_id]['tracesInfo'].
        tracesInfo includes list of traces' info dict which is sorted in
            startTmEpoch order by sort_data.
            refer to DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id]['traces_info'].

        """
        count = 0
        for DS in self.waveformData[key]['filesInfo']:
            readData = self.waveformData[key]['readData']
            for fileInfo in self.waveformData[key]['filesInfo'][DS]:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()
                # file have been read
                if fileInfo['read']:
                    continue

                # check time
                hasData = False
                if ((self.readStart <= fileInfo['startEpoch'] <= self.readEnd) or   # noqa: E501
                        (self.readStart <= fileInfo['endEpoch'] <= self.readEnd)):  # noqa: E501
                    hasData = True
                if not hasData:
                    continue
                readWaveformReftek(fileInfo['rt130'], key, readData,
                                   self.dataTime[key], self.tmpDir)
                fileInfo['read'] = True
                count += 1
                if count % 50 == 0:
                    self.trackInfo(
                        f'Read {count} waveform files', LogType.INFO)
        sortData(self.waveformData)

    def addLog(self, chan_pkt, logInfo):
        if chan_pkt not in self.logData[self.curKey].keys():
            self.logData[self.curKey][chan_pkt] = []
        self.logData[self.curKey][chan_pkt].append(logInfo)

    def readReftek130(self, path2file):
        """
        From the given file:
            + Read SOH data from file with SH packets,
            + read event info, mass position and index waveform (data stream)
                file from file with EH or ET packets
        :param path2file: str - absolute path to file
        """
        rt130 = core.Reftek130.from_file(path2file)
        unique, counts = np.unique(rt130._data["packet_type"],
                                   return_counts=True)
        nbr_packet_type = dict(zip(unique, counts))

        if b"SH" in nbr_packet_type:
            self.readSH(path2file)
        if b"EH" in nbr_packet_type or b"ET" in nbr_packet_type:
            self.readEHET_MP_indexWF(rt130)
        return True

    def readSH(self, path2file):
        """
        Use soh_packet library to read file with SH packet for soh data
            to append tuple (time, log string) to log_data[self.curKey][SOH]
        :param path2file: str - absolute path to file
        """
        with open(path2file, "rb") as fh:
            str = fh.read()
        data = soh_packet._initial_unpack_packets_soh(str)
        for ind, d in enumerate(data):
            self.curKey = (d['unit_id'].decode(),
                           d['experiment_number'])
            if self.curKey not in self.logData:
                self.logData[self.curKey] = {}
            logs = soh_packet.Packet.from_data(d).__str__()
            # self.addLog(d['packet_type'].decode(), (d['time'], logs))
            if 'SOH' not in self.logData[self.curKey]:
                self.logData[self.curKey]['SOH'] = []
            self.logData[self.curKey]['SOH'].append((d['time'], logs))

    def readEHET_MP_indexWF(self, rt130):
        """
        Files that contents EH or ET packets are data stream files.
        There may be 1 - 9 data streams.
        :param rt130: rt130 object - of a data stream file in which
            + event info can be found in EH packet and save in self.logData
            + mass position data can be found in data stream 9 and save in
                self.masspos_stream
            + waveform data can be found in data stream 1-8 which will be index
                and save to self.waveformData[self.curKey]["filesInfo"] to
                be processed later
        """
        DS = rt130._data['data_stream_number'][0] + 1
        if DS not in self.reqDSs + [9]:
            return

        ind_EHET = [ind for ind, val in
                    enumerate(rt130._data["packet_type"])
                    if val in [b"EH"]]     # only need event header
        nbr_DT_samples = sum(
            [rt130._data[ind]["number_of_samples"]
             for ind in range(0, len(rt130._data))
             if ind not in ind_EHET])

        for ind in ind_EHET:
            d = rt130._data[ind]
            self.curKey = (d['unit_id'].decode(),
                           d['experiment_number'])
            if self.curKey not in self.logData:
                self.logData[self.curKey] = {}
            logs = packet.EHPacket(d).eh_et_info(nbr_DT_samples)
            if 'EHET' not in self.logData[self.curKey]:
                self.logData[self.curKey]['EHET'] = []
            self.logData[self.curKey]['EHET'].append((d['time'], logs))
        if self.curKey not in self.dataTime:
            self.dataTime[self.curKey] = [constants.HIGHEST_INT, 0]
            self.keys.add(self.curKey)
        if DS == 9:
            self.readMassPos(rt130)
        else:
            self.indexWaveForm(rt130, DS)

    def readMassPos(self, rt130):
        """
        Append all traces of data stream 9 to self.massPosStream[self.currKey].
        Update dataTime.
        :param rt130: object rt130 - for data stream 9
        """
        if self.curKey not in self.massPosStream:
            self.massPosStream[self.curKey] = Stream()

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=False,
            verbose=False,
            sort_permuted_package_sequence=True)
        for tr in stream:
            self.massPosStream[self.curKey].append(tr)
            self.dataTime[self.curKey][0] = min(
                tr.stats['starttime'].timestamp, self.dataTime[self.curKey][0])
            self.dataTime[self.curKey][1] = max(
                tr.stats['endtime'].timestamp, self.dataTime[self.curKey][1])

    def indexWaveForm(self, rt130, DS):
        """
        Indexing by adding rt130 object along with time range to
            self.waveformData[self.currKey]['filesInfo']
        :param rt130: object rt130 - for data stream DS
        :param DS: int - data stream index
        """
        if self.curKey not in self.waveformData:
            self.waveformData[self.curKey] = {"filesInfo": {},
                                              "readData": {}}
        if DS not in self.waveformData[self.curKey]["filesInfo"]:
            self.waveformData[self.curKey]["filesInfo"][DS] = []

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=True,
            verbose=False,
            sort_permuted_package_sequence=True)

        self.waveformData[self.curKey]["filesInfo"][DS].append(
            {'rt130': rt130,
             'startEpoch': stream[0].stats['starttime'].timestamp,
             'endEpoch': stream[0].stats['endtime'].timestamp,
             'read': False})

    def combineData(self):
        """
        + SOH and event logData will be processed to create SOHdata:
            {Key of set:
                {chan_id:
                    {orgTrace: {
                        times: list of epoch time
                        data: list of channel's values
                        logIdx: list of indexes of SOH message line
                            corresponding to time and data
                    }}
                }
            }
        + massPossStream for each data set will be merged and save to
        massPossData according to chan_ids
        + Gaps will be calculate and squash for mass position only since
        currently there is no way to calculate SOH gaps, and data for waveform
        is too big to consider calculating gaps.
        """
        for k in self.logData:
            if self.creator_thread.isInterruptionRequested():
                raise ThreadStopped()
            if k == 'TEXT':
                continue
            if k not in self.dataTime:
                self.dataTime[k] = [constants.HIGHEST_INT, 0]
                self.keys.add(k)
            self.SOHData[k] = {}
            logs = []
            for pktType in ['SOH', 'EHET']:
                if pktType == 'EHET':
                    logs += '\n\nEvents:'
                try:
                    for log in sorted(self.logData[k][pktType],
                                      key=lambda x: x[0]):
                        # sort log data according to time
                        logs += log[1]
                        if pktType == 'SOH':
                            logs += '\n'
                except KeyError:
                    pass
            logStr = ''.join(logs)
            self.logData[k] = {'SOH': [logStr]}
            logObj = LogInfo(
                self, self.trackInfo, logStr, k, self.reqDSs)
            self.dataTime[k][0] = min(logObj.minEpoch, self.dataTime[k][0])
            self.dataTime[k][1] = max(logObj.maxEpoch, self.dataTime[k][1])
            for cName in self.SOHData[k]:
                c = self.SOHData[k][cName]['orgTrace']
                c['times'] = np.array(c['times'])
                c['data'] = np.array(c['data'])
                c['logIdx'] = np.array(c['logIdx'])

        allGaps = []
        self.gaps = {k: [] for k in self.keys}
        self.massPosData = {k: {} for k in self.keys}
        for k in self.massPosStream:
            if self.creator_thread.isInterruptionRequested():
                raise ThreadStopped()
            stream = self.massPosStream[k]
            stream.merge()
            for tr in stream:
                chanID = tr.stats['channel']
                self.massPosData[k][chanID] = {}
                self.massPosData[k][chanID]['orgTrace'] = readMPTrace(tr)
            gaps = stream.get_gaps()
            # gaps is list of [network, station, location, channel, starttime,
            # endtime, duration, number of missing samples]
            allGaps += [[g[4].timestamp, g[5].timestamp] for g in gaps]
            self.gaps[k] = squash_gaps(allGaps)
