"""
Drawing State-Of-Health channels and mass position
"""
from sohstationviewer.view.util.plot_func_names import plot_functions
from sohstationviewer.view.plotting.plotting_widget import plotting_widget

from sohstationviewer.controller.plottingData import getTitle
from sohstationviewer.controller.util import (
    displayTrackingInfo, apply_convert_factor)

from sohstationviewer.conf import constants

from sohstationviewer.database import extractData

from sohstationviewer.model.handling_data import trim_downsample_SOHChan

from sohstationviewer.view.util.enums import LogType


class SOHWidget(plotting_widget.PlottingWidget):
    def plot_channels(self, start_tm, end_tm, key, data_time,
                      channel_list, time_ticks_total,
                      soh_data, mass_pos_data, gaps):
        """
        Recursively plot each channels for soh_data and masspos_data.

        :param start_tm: float - requested start time to read
        :param end_tm: float - requested end time to read
        :param key: str or (str, str) station name  for mseed,
            or (device's experiment number, experiment number ) for reftek
        :param data_time: [float, float] - time range of the selected data set
        :param channel_list: [str,] - SOH channel list read from the selected
            data set
        :param time_ticks_total: int - max number of tick to show on time bar
        :param soh_data: dict - SOH data of the selected data set, refer to
            DataTypeModel.__init__.soh_data[key]
        :param mass_pos_data: dict - mass position data of the selected
            data set, refer to DataTypeModel.__init__.mass_pos_data[key]
         :param gaps: [[float, float],] - list of gaps of the selected data set
        """
        self.plotting_data1 = soh_data
        self.plotting_data2 = mass_pos_data
        self.processing_log = []
        self.gap_bar = None

        if self.axes:
            self.plotting_axes.fig.clear()
        self.date_mode = self.parent.date_format.upper()
        self.time_ticks_total = time_ticks_total

        self.min_x = max(data_time[0], start_tm)
        self.max_x = min(data_time[1], end_tm)
        self.plot_total = len(self.plotting_data1) + len(self.plotting_data2)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = constants.BOTTOM
        self.plotting_bot_pixel = constants.BOTTOM_PX
        self.axes = []

        self.timestamp_bar_top = self.plotting_axes.add_timestamp_bar(0.003)
        self.plotting_axes.set_title(title)
        self.plotting_axes.add_gap_bar(gaps)
        not_found_chan = [c for c in channel_list
                          if c not in self.plotting_data1.keys()]
        if len(not_found_chan) > 0:
            msg = (f"The following channels is in Channel Preferences but "
                   f"not in the given data: {not_found_chan}")
            self.processing_log.append((msg, LogType.WARNING))

        for chan_id in self.plotting_data1:
            chan_db_info = extractData.getChanPlotInfo(chan_id,
                                                       self.parent.data_type)
            if chan_db_info['height'] == 0:
                # not draw
                continue
            if chan_db_info['channel'] == 'DEFAULT':
                msg = (f"Channel {chan_id}'s "
                       f"definition can't be found database.")
                displayTrackingInfo(self.tracking_box, msg, LogType.WARNING)

            if chan_db_info['plotType'] == '':
                continue
            self.plotting_data1[chan_id]['chan_db_info'] = chan_db_info
            self.get_zoom_data(self.plotting_data1[chan_id], chan_id,
                               self.plotting_data1, True)

        for chan_id in self.plotting_data2:
            chan_db_info = extractData.getChanPlotInfo(
                chan_id, self.parent.data_type)
            self.plotting_data2[chan_id]['chan_db_info'] = chan_db_info
            self.get_zoom_data(self.plotting_data2[chan_id], chan_id, True)

        self.axes.append(self.plotting.plot_none())
        self.timestamp_bar_bottom = self.plotting_axes.add_timestamp_bar(
            0.003, top=False)
        self.set_lim(first_time=True)
        self.bottom = self.axes[-1].get_ybound()[0]
        self.ruler = self.plotting_axes.add_ruler(
            self.display_color['time_ruler'])
        self.zoom_marker1 = self.plotting_axes.add_ruler(
            self.display_color['zoom_marker'])
        self.zoom_marker2 = self.plotting_axes.add_ruler(
            self.display_color['zoom_marker'])
        # Set view size fit with the given data
        if self.main_widget.geometry().height() < self.plotting_bot_pixel:
            self.main_widget.setFixedHeight(self.plotting_bot_pixel)

        self.draw()

    def get_zoom_data(self, c_data, chan_id, plotting_data=None,
                      first_time=False):
        """
        Trim data of a channel to the zooming time range and re-downsample
            the data to get more detail.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to
            DataTypeModel.__init__.soh_data[key][chan_id] or
            DataTypeModel.__init__.mass_pos_data[key][chan_id]
        :param chan_id: str - name of channel
        :param plotting_data: dict - DataTypeModel.__init__.soh_data[key]. To
            help combine 2 channels in one axes in need
        :param first_time: bool - flag identify when the data set is first
            plotted, not zoom in yet
        """
        chan_db_info = c_data['chan_db_info']
        plot_type = chan_db_info['plotType']
        trim_downsample_SOHChan(c_data, self.min_x, self.max_x,
                                first_time)
        apply_convert_factor(c_data, 1)
        if 'ax' not in c_data:
            linked_ax = None
            if chan_db_info['linkedChan'] not in [None, 'None', '']:
                linked_ax = plotting_data[chan_db_info['linkedChan']]['ax']

            ax = getattr(self.plotting, plot_functions[plot_type][1])(
                c_data, chan_db_info, chan_id, None, linked_ax)
            if ax is None:
                return
            c_data['ax'] = ax
            ax.chan = chan_id
            self.axes.append(ax)
        else:
            getattr(self.plotting, plot_functions[plot_type][1])(
                c_data, chan_db_info, chan_id, c_data['ax'], None)
